@@ -1,7 +1,9 @@
+import json
 from sys import stderr
 from time import sleep
-from typing import List
+from typing import List, Dict, Optional
 
+import jupytext
 from loguru import logger
 from selenium import webdriver
 from selenium.common.exceptions import NoSuchElementException, WebDriverException
@@ -10,7 +12,12 @@ from watchdog.events import FileSystemEventHandler
 
 from jure.events import reload_event, cells_changed_event, EventType
 from jure.index_notebook_cells import CellsIndex
-from jure.utils import get_notebook_path_from_file_path, get_lines_file, get_file_update_timestamp, get_diffing_lines
+from jure.utils import (
+    get_notebook_path_from_file_path,
+    get_file_update_timestamp,
+    get_diffing_lines,
+    get_file,
+)
 
 logger.remove()
 logger.add(stderr, format="{time: HH:mm:SSSS} {level} {message}", level="INFO")
@@ -27,25 +34,26 @@ class BaseHandler:
 class SeleniumHandler(BaseHandler):
     def __init__(self, page):
         options = Options()
-        options.add_experimental_option('excludeSwitches', ['enable-automation'])
-        options.add_experimental_option('useAutomationExtension', False)
-        options.set_capability('unhandledPromptBehaviour', 'accept')
+        options.add_experimental_option("excludeSwitches", ["enable-automation"])
+        options.add_experimental_option("useAutomationExtension", False)
+        options.set_capability("unhandledPromptBehaviour", "accept")
         options.add_argument("--disable-popup-blocking")
         self.driver = webdriver.Chrome(options=options)
         self.driver.execute_script("window.onbeforeunload = null;")
         try:
             self.driver.get(page)
+            self.driver.execute_script("Jupyter.notebook.set_autosave_interval(0);")
         except WebDriverException as e:
             msg = f"""Unable to load Jupyter Notebook.
                       Make sure that Jupyter Notebook is available on page {page}"""
             raise RuntimeError(msg)
 
     def handle(self, event):
-        if event['type'] == EventType.RELOAD_PAGE:
+        if event["type"] == EventType.RELOAD_PAGE:
             self._refresh_page()
-        if event['type'] == EventType.HANDLE_CHANGED_CELLS:
-            self._scroll_to_cell(event['cell_to_scroll'])
-            self._execute_cells(event['changed_cells'])
+        if event["type"] == EventType.HANDLE_CHANGED_CELLS:
+            self._scroll_to_cell(event["cell_to_scroll"])
+            self._execute_cells(event["changed_cells"])
 
     def shutdown(self):
         self.driver.close()
@@ -55,20 +63,28 @@ class SeleniumHandler(BaseHandler):
         sleep(0.1)
         self.driver.refresh()
 
-    def _scroll_to_cell(self, cell_num):
+    def _scroll_to_cell(self, cell_num: Optional[int]):
         self.check_popup()
-        logger.info(f"Scrolling to cell {cell_num}")
-        self.driver.execute_script(
-            f"""
-            var cell = Jupyter.notebook.get_cell({cell_num});
-            cell.element[0].scrollIntoView();
-            """)
+        if cell_num is not None:
+            logger.info(f"Scrolling to cell {cell_num}")
+            self.driver.execute_script(
+                f"""
+                Jupyter.notebook.scroll_to_cell({cell_num});
+                """
+            )
 
-    def _execute_cells(self, changed_cells: List[str]):
-        logger.info(f"Executing cells {changed_cells}")
+    def _execute_cells(self, changed_cells: List[Dict]):
+        logger.info(f"Executing")
         if changed_cells:
+            formatted_cells = [json.dumps(c, ensure_ascii=False) for c in changed_cells]
+            update_script = f"""let data = [{','.join(formatted_cells)}];
+data.forEach(function (cell) {{
+    Jupyter.notebook.get_cell(cell.index).set_text(cell.content)
+}});
+"""
+            self.driver.execute_script(update_script)
             exec_script = f"""
-                Jupyter.notebook.execute_cells([{','.join(changed_cells)}]);
+                Jupyter.notebook.execute_cells([{','.join([c["index"] for c in changed_cells])}]);
                 """
             self.driver.execute_script(exec_script)
 
@@ -79,8 +95,12 @@ class SeleniumHandler(BaseHandler):
             pass
         reload_btn = None
         try:
-            notebook_changed_modal = self.driver.find_element_by_css_selector("body.modal-open")
-            reload_btn = notebook_changed_modal.find_element_by_class_name("btn-warning")
+            notebook_changed_modal = self.driver.find_element_by_css_selector(
+                "body.modal-open"
+            )
+            reload_btn = notebook_changed_modal.find_element_by_class_name(
+                "btn-warning"
+            )
         except NoSuchElementException:
             pass
         if reload_btn is not None:
@@ -92,38 +112,42 @@ class SeleniumHandler(BaseHandler):
 
 
 class WatchdogHandler(FileSystemEventHandler):
-
     def __init__(self, file_path, handler):
         self.file_path = file_path
         self.notebook_path = get_notebook_path_from_file_path(file_path)
-        self.file_lines = get_lines_file(self.file_path)
+        self.prev_cells = jupytext.reads(get_file(self.file_path), "py:percent")[
+            "cells"
+        ]
         self.prev_update_timestamp = 0
         self.handler = handler
 
     def on_modified(self, event):
-        logger.info(f'event type: {event}')
+        logger.info(f"event type: {event}")
         file_update_timestamp = get_file_update_timestamp(self.file_path)
         notebook_update_timestamp = get_file_update_timestamp(self.notebook_path)
         if self.should_reload(file_update_timestamp, notebook_update_timestamp):
-            logger.info("Reloading")
+            logger.info("Updating notebook text")
             try:
-                self.handler.handle(reload_event)
-                new_lines = get_lines_file(self.file_path)
-                diffing_lines = get_diffing_lines(self.file_lines, new_lines)
-                self.file_lines = new_lines
-                cells_index = CellsIndex(new_lines)
-                if not diffing_lines:
-                    cell_to_scroll = cells_index.get_last_cell()
-                else:
-                    cell_to_scroll = cells_index.get_cell(diffing_lines[-1])
-                diffing_cells = set()
-                for line in diffing_lines:
-                    diffing_cells.add(cells_index.get_cell(line))
-                diffing_cells = [str(cell) for cell in sorted(diffing_cells)]
-                new_scroll_event = cells_changed_event(cell_to_scroll=cell_to_scroll,
-                                                       diffing_cells=diffing_cells)
-                logger.info(str(new_scroll_event))
-                self.handler.handle(new_scroll_event)
+                cells = jupytext.reads(get_file(self.file_path), "py:percent")["cells"]
+                diffing_cells = [
+                    i
+                    for i, cell in enumerate(cells)
+                    if i >= len(self.prev_cells)
+                    or self.prev_cells[i]["source"] != cell["source"]
+                ]
+                self.prev_cells = cells
+                cell_to_scroll = None
+                if diffing_cells:
+                    cell_to_scroll = diffing_cells[-1]
+                cell_info = [
+                    {"index": str(cell_index), "content": cells[cell_index]["source"]}
+                    for cell_index in sorted(diffing_cells)
+                ]
+                new_event = cells_changed_event(
+                    cell_to_scroll=cell_to_scroll, diffing_cells=cell_info
+                )
+                logger.info(f"{new_event['type']}. Changed cells: {diffing_cells}")
+                self.handler.handle(new_event)
             except Exception as e:
                 logger.exception(e)
         self.prev_update_timestamp = file_update_timestamp
